import enum
from typing import Union

@enum.unique
class SetVerboseMode(enum.Enum):
  OFF = "0"
  INFO = "2"
  VERBOSE = "3"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetHdmiMode(enum.Enum):
  AUTO = "AUTO"
  SRC = "SRC"
  UHD_AUTO = "UHD_AUTO"
  UHD24 = "UHD24"
  UHD50 = "UHD50"
  UHD60 = "UHD60"
  HD1080P_AUTO = "1080P_AUTO"
  HD1080P24 = "1080P24"
  HD1080P50 = "1080P50"
  HD1080P60 = "1080P60"
  HD1080I50 = "1080I50"
  HD1080I60 = "1080I60"
  HD720P50 = "720P50"
  HD720P60 = "720P60"
  SD576P = "576P"
  SD576I = "576I"
  SD480P = "480P"
  SD480I = "480I"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetZoomMode(enum.Enum):
  Z1 = "1"
  AR = "AR"
  FS = "FS"
  US = "US"
  Z1P2 = "1.2"
  Z1P3 = "1.3"
  Z1P5 = "1.5"
  Z2 = "2"
  Z0P2 = "1/2"
  Z3 = "3"
  Z4 = "4"
  Z0P3 = "1/3"
  Z0P4 = "1/4"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetVolumeLevel(enum.Enum):
  MUTE = "MUTE"

  def __str__(self):
    return str(self.value)

VolumeLevelType = Union[SetVolumeLevel, int]

@enum.unique
class SetRepeatMode(enum.Enum):
  CHAPTER = "CH"
  TRACK = "TT"
  ALL = "ALL"
  OFF = "OFF"
  SHUFFLE = "SHF"
  RANDOM = "RND"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetSearchMode(enum.Enum):
  CHAPTER = "C"
  TITLE = "T"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetTimeCodeMode(enum.Enum):
  TOTAL_ELAPSED = "E"
  TOTAL_REMAINING = "R"
  TITLE_ELAPSED = "T"
  TITLE_REMAINING = "X"
  CHAPTER_ELAPSED = "C"
  CHAPTER_REMAINING = "K"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetHdrMode(enum.Enum):
  AUTO = "Auto"
  ON = "On"
  OFF = "Off"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetInputSource(enum.Enum):
  BLURAY = "0"
  HDMI_IN = "1"
  HDMI_ARC = "2"
  OPTICAL_IN = "3"
  COAX_IN = "4"
  USB_IN = "5"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetScreenSaverMode(enum.Enum):
  ON = "ON"
  OFF = "OFF"
  SAVE = "SAVE"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetAppMode(enum.Enum):
  DISC = "DIS"
  MUSIC = "MUS"
  PHOTO = "PHO"
  MOVIE = "MOV"
  NETWORK = "NET"
  SETUP = "SET"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetSacdPriority(enum.Enum):
  MULTI_CHANNEL = "M"
  STEREO = "S"
  CD = "C"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetSacdOutputMode(enum.Enum):
  DSD = "D"
  PCM = "P"
  AUTO = "A"

  def __str__(self):
    return str(self.value)

@enum.unique
class SetSpeedMode(enum.Enum):
  SLOW_1_32 = "1/32"
  SLOW_1_16 = "1/16"
  SLOW_1_8 = "1/8"
  SLOW_1_4 = "1/4"
  SLOW_1_2 = "1/2"
  NORMAL = "1"
  FAST_2 = "2"
  FAST_3 = "3"
  FAST_4 = "4"
  FAST_5 = "5"

  def __str__(self):
    return str(self.value)

  


MAX_RETRIES = 3
RETRY_INTERVAL = 2

#occurs when the client is connected
EVENT_CONNECTED = "connected"
#occurs when the client is disconnected
EVENT_DISCONNECTED = "disconnected"
#occurs when the client state has changed
EVENT_STATE_CHANGED = "state_changed"
#occurs when the client is ready
EVENT_READY = "ready"
#occurs when the client receives a message
EVENT_MESSAGE_RECEIVED = "message_received"
#occurs when a command is about to be sent to the device
EVENT_COMMAND_SENDING = "command_sending"
#occurs when a command has been sent to the device
EVENT_COMMAND_SENT = "command_sent"
#occurs when the command response has been received
EVENT_COMMAND_RESPONSE = "command_response"
#occurs when the command processing has been completed
EVENT_COMMAND_COMPLETE = "command_complete"
#occurs when the device is updating
EVENT_DEVICE_STATE_UPDATING = "device_state_updating"
#occurs after the device state has been updated
EVENT_DEVICE_STATE_UPDATED = "device_state_updated"

#per Oppo documentation, assume 10 second timeout
COMMAND_TIMEOUT = 10

ATTR_DEVICE_FIRMWARE_VERSION = "firmware_version"
ATTR_DEVICE_IS_MUTED = "is_muted"
ATTR_DEVICE_VOLUME = "volume"
ATTR_DEVICE_HDMI_MODE = "hdmi_mode"
ATTR_DEVICE_SUBTITLE_SHIFT = "subtitle_shift"
ATTR_DEVICE_OSD_POSITION = "osd_position"
ATTR_DEVICE_INPUT_SOURCE = "input_source"
ATTR_DEVICE_ZOOM_MODE = "zoom_mode"
ATTR_DEVICE_HDR_SETTING = "hdr_setting"
ATTR_DEVICE_DISC_TYPE = "disc_type"
ATTR_DEVICE_PLAYBACK_STATUS = "playback_status"
ATTR_DEVICE_POWER_STATUS = "power_status"
ATTR_DEVICE_CDDB_ID_1 = 'cddb_id_1'
ATTR_DEVICE_CDDB_ID_2 = 'cddb_id_2'
ATTR_DEVICE_CDDB_ID = "cddb_id"

ATTR_PLAYBACK_TRACK = "track"
ATTR_PLAYBACK_TRACK_TOTAL = "track_total"
ATTR_PLAYBACK_CHAPTER = "chapter"
ATTR_PLAYBACK_CHAPTER_TOTAL = "chapter_total"
ATTR_PLAYBACK_TRACK_ELAPSED_TIME = "track_elapsed_time"
ATTR_PLAYBACK_TRACK_REMAINING_TIME = "track_remaining_time"
ATTR_PLAYBACK_TRACK_DURATION = "track_duration"
ATTR_PLAYBACK_CHAPTER_ELAPSED_TIME = "chapter_elapsed_time"
ATTR_PLAYBACK_CHAPTER_REMAINING_TIME = "chapter_remaining_time"
ATTR_PLAYBACK_CHAPTER_DURATION = "chapter_duration"
ATTR_PLAYBACK_TOTAL_ELAPSED_TIME = "total_elapsed_time"
ATTR_PLAYBACK_TOTAL_REMAINING_TIME = "total_remaining_time"
ATTR_PLAYBACK_TOTAL_DURATION = "total_duration"
ATTR_PLAYBACK_AUDIO_TYPE = "audio_type"
ATTR_PLAYBACK_SUBTITLE_TYPE = "subtitle_type"
ATTR_PLAYBACK_ASPECT_RATIO = "aspect_ratio"
ATTR_PLAYBACK_REPEAT_MODE = "repeat_mode"
ATTR_PLAYBACK_VIDEO_3D_STATUS = "video_3d_status"
ATTR_PLAYBACK_VIDEO_HDR_STATUS = "video_hdr_status"
ATTR_PLAYBACK_MEDIA_FILE_FORMAT = "media_file_format"
ATTR_PLAYBACK_MEDIA_FILE_NAME = "media_file_name"
ATTR_PLAYBACK_TRACK_NAME = "track_name"
ATTR_PLAYBACK_TRACK_ALBUM = "track_album"
ATTR_PLAYBACK_TRACK_PERFORMER = "track_performer"
ATTR_PLAYBACK_REV_SPEED = "rev_speed"
ATTR_PLAYBACK_FWD_SPEED = "fwd_speed"
ATTR_PLAYBACK_CAMERA_ANGLE = "camera_angle"
ATTR_PLAYBACK_ZOOM_RATIO = "zoom_ratio"
ATTR_PLAYBACK_ATB_REPLAY = "atb_replay"


import logging
from datetime import datetime, timedelta
from dataclasses import dataclass
from typing import List, NamedTuple, Optional, TYPE_CHECKING

from ..codes import *
from .enums import *
from .mutator import OppoNopMutator, OppoStateMutator
from .mapping import (
  _UPDATE_PLAY_STATUS_TO_PLAY_STATUS, 
  _UPDATE_PLAY_STATUS_TO_SPEED_MODE, 
  _UPDATE_PLAY_STATUS_TO_TRAY_STATUS,
  _UPDATE_DISC_TYPE_TO_DISC_TYPE
)

if TYPE_CHECKING:
    from ..device import OppoDevice

_LOGGER = logging.getLogger(__name__)    

class OppoParsedResponse(NamedTuple):
  code: str
  result: str
  parameters: List[str] = None

@dataclass(repr=True, eq=True)
class OppoResponse:
  """Represents a response to a command or state update"""
  code: OppoCodeType
  result: ResultCode
  _parameters: List[str]
  _raw_value: Optional[str]
  
  def __init__(self, parsed: OppoParsedResponse, mutator: OppoStateMutator = OppoNopMutator(), raw_value: Optional[str] = None, single_parameter: bool = True):
    self.code = self._translate(parsed.code)
    self.result = ResultCode(parsed.result)
    
    #note: many responses can have spaces, so by default
    #we'll just rejoin all the rest of the parameters
    #if we need to split again later, we can do that in a subclass
    if single_parameter:
      self._parameters = [" ".join(parsed.parameters)]
    else:
      self._parameters = parsed.parameters
    self._raw_value = raw_value
    self._mutator = mutator

  @property
  def raw_value(self):
    return self._raw_value

  def _translate(self, code: OppoCodeType):
    try:
      if isinstance(code, str):
        return OppoCode(code)    
    except:
      _LOGGER.info(f'Unexpected code received: {code}')
      pass
    return code

  async def mutate_state(self, device: 'OppoDevice'):
    await self._mutator.mutate_state(device, self)

class OppoStringResponse(OppoResponse):
  @property
  def value(self) -> str:
    return self._parameters[0]

class OppoIntResponse(OppoResponse):
  @property
  def value(self) -> str:
    return int(self._parameters[0])

class OppoTimeResponse(OppoResponse):
  @property
  def value(self) -> timedelta:
    t = datetime.strptime(self._parameters[0],"%H:%M:%S")
    return timedelta(hours=t.hour,minutes=t.minute,seconds=t.second)

class OppoPowerResponse(OppoResponse):
  @property
  def status(self) -> PowerStatus:
    return PowerStatus(self._parameters[0])

class OppoPlayResponse(OppoResponse):
  @property
  def status(self) -> PlayStatus:
    return PlayStatus(self._parameters[0])

class OppoHdmiModeResponse(OppoResponse):
  @property
  def mode(self) -> PlayStatus:
    return HdmiMode(self._parameters[0])

class OppoVolumeLevelResponse(OppoResponse):
  @property
  def level(self) -> VolumeLevelType:
    try:
      return int(self._parameters[0])
    except:
      return VolumeLevel(self._parameters[0])

class OppoZoomModeResponse(OppoResponse):
  def __init__(self, parsed: OppoParsedResponse, mutator: OppoStateMutator, raw_value: Optional[str] = None):
      super().__init__(parsed, mutator, raw_value, False)
  @property
  def mode(self) -> ZoomMode:
    return ZoomMode(self._parameters[0])

class OppoInputSourceResponse(OppoResponse):
  def __init__(self, parsed: OppoParsedResponse, mutator: OppoStateMutator, raw_value: Optional[str] = None):
      super().__init__(parsed, mutator, raw_value, False)
  @property
  def source(self) -> InputSource:
    return InputSource(self._parameters[0])

class OppoDiscTypeResponse(OppoResponse):
  @property
  def disc_type(self) -> DiscType:
    return DiscType(self._parameters[0])

class OppoHdrSettingResponse(OppoResponse):
  @property
  def setting(self) -> HdrSetting:
    return HdrSetting(self._parameters[0])

class OppoRepeatModeResponse(OppoResponse):
  def __init__(self, parsed: OppoParsedResponse, mutator: OppoStateMutator, raw_value: Optional[str] = None):
      super().__init__(parsed, mutator, raw_value, False)
  @property
  def mode(self) -> RepeatMode:
    return RepeatMode(self._parameters[0])

class OppoVideo3dStatusResponse(OppoResponse):
  @property
  def status(self) -> Video3dStatus:
    return Video3dStatus(self._parameters[0])

class OppoVideoHdrStatusResponse(OppoResponse):
  @property
  def status(self) -> VideoHdrStatus:
    return VideoHdrStatus(self._parameters[0])

class OppoSpeedModeResponse(OppoResponse):
  @property
  def mode(self) -> SpeedMode:
    return SpeedMode(self._parameters[0])

class OppoTrayStatusResponse(OppoResponse):
  @property
  def status(self) -> TrayStatus:
    return TrayStatus(self._parameters[0])    

class OppoCurrentTotalResponse(OppoResponse):
  @property
  def current(self) -> int:
    return int(self._parameters[0].split('/')[0])  
  @property
  def total(self) -> int:
    return int(self._parameters[0].split('/')[1])

class OppoUpdatePowerStatusResponse(OppoResponse):
  @property
  def status(self) -> PowerStatus:
    return PowerStatus.ON if int(self._parameters[0]) == 1 else PowerStatus.OFF

class OppoUpdatePlayStatusResponse(OppoResponse):
  @property
  def update_status(self) -> UpdatePlayStatus:
    return UpdatePlayStatus(self._parameters[0])
  @property
  def play_status(self) -> Optional[PlayStatus]:
    try:
      return _UPDATE_PLAY_STATUS_TO_PLAY_STATUS[self.update_status]
    except:
      return None
  @property
  def tray_status(self) -> Optional[TrayStatus]:
    try:
      return _UPDATE_PLAY_STATUS_TO_TRAY_STATUS[self.update_status]
    except:
      return None
  @property
  def fwd_speed_mode(self) -> Optional[SpeedMode]:
    try:
      if 'FW' in self.update_status.value:
        return _UPDATE_PLAY_STATUS_TO_SPEED_MODE[self.update_status]
    except:
      pass
    return None
  @property
  def rev_speed_mode(self) -> Optional[SpeedMode]:
    try:
      if 'RV' in self.update_status.value:
        return _UPDATE_PLAY_STATUS_TO_SPEED_MODE[self.update_status]
    except:
      pass
    return None 

class OppoUpdateDiscTypeResponse(OppoResponse):
  @property
  def disc_type(self) -> DiscType:
    return _UPDATE_DISC_TYPE_TO_DISC_TYPE[UpdateDiscType(self._parameters[0])]

class OppoUpdateTimeResponse(OppoResponse):
  def __init__(self, parsed: OppoParsedResponse, mutator: OppoStateMutator, raw_value: Optional[str] = None):
      super().__init__(parsed, mutator, raw_value, False)  
  @property
  def title_number(self):
    return int(self._parameters[0])
  @property
  def chapter_number(self):
    return int(self._parameters[1])
  @property
  def time_type(self):
    return self._parameters[2]
  @property
  def time_value(self) -> timedelta:
    t = datetime.strptime(self._parameters[3],"%H:%M:%S")
    return timedelta(hours=t.hour,minutes=t.minute,seconds=t.second)  


from .enums import *
from .response import OppoResponse
from .response import OppoIntResponse
from .response import OppoStringResponse
from .response import OppoTimeResponse
from .response import OppoPowerResponse
from .response import OppoPlayResponse
from .response import OppoDiscTypeResponse
from .response import OppoHdmiModeResponse
from .response import OppoHdrSettingResponse
from .response import OppoInputSourceResponse
from .response import OppoRepeatModeResponse
from .response import OppoSpeedModeResponse
from .response import OppoVideoHdrStatusResponse
from .response import OppoVideo3dStatusResponse
from .response import OppoZoomModeResponse
from .response import OppoUpdatePowerStatusResponse
from .response import OppoUpdatePlayStatusResponse
from .response import OppoUpdateTimeResponse
from .factory import get_response


from typing import List, Optional
from datetime import datetime, timedelta
from ..codes import OppoSetCode, OppoSetCodeType, OppoCode, OppoCodeType
from ..helpers import *
from .enums import *
from .command import OppoCommand

class OppoSetCommand(OppoCommand):
  def __init__(self, code: OppoSetCodeType, parameters: List[str] = None):      
    super().__init__(code, parameters)

  def _translate(self, code: OppoSetCodeType):
    if isinstance(code, str):
      return OppoCode(OppoSetCode(code).value)
    return OppoCode(code.value)

class OppoSetVerboseModeCommand(OppoSetCommand):
  def __init__(self, mode: SetVerboseMode):
    super().__init__(OppoSetCode.SVM)
    self._parameters.append(mode)

class OppoSetHdmiModeCommand(OppoSetCommand):
  def __init__(self, mode: SetHdmiMode):
    super().__init__(OppoSetCode.SHD)
    self._parameters.append(mode)

class OppoSetZoomModeCommand(OppoSetCommand):
  def __init__(self, mode: SetZoomMode):
    super().__init__(OppoSetCode.SZM)
    self._parameters.append(mode)

class OppoSetVolumeLevelCommand(OppoSetCommand):
  def __init__(self, level: VolumeLevelType):
    super().__init__(OppoSetCode.SVL)
    
    if isinstance(level, int):
      self._parameters.append(str(clamp(level,0,100)))
    else:
      self._parameters.append(level)

class OppoSetRepeatModeCommand(OppoSetCommand):
  def __init__(self, mode: SetRepeatMode):
    super().__init__(OppoSetCode.SRP)
    self._parameters.append(mode)

class OppoSetSubtitleShiftCommand(OppoSetCommand):
  def __init__(self, shift: int):
    super().__init__(OppoSetCode.SSH)
    self._parameters.append(str(clamp(shift,-10,10)))

class OppoSetOsdPositionCommand(OppoSetCommand):
  def __init__(self, shift: int):
    super().__init__(OppoSetCode.SOP)
    self._parameters.append(str(clamp(shift,0,5)))

class OppoSetTimeCodeModeCommand(OppoSetCommand):
  def __init__(self, mode: SetTimeCodeMode):
    super().__init__(OppoSetCode.STC)
    self._parameters.append(mode)

class OppoSetHdrModeCommand(OppoSetCommand):
  def __init__(self, mode: SetHdrMode):
    super().__init__(OppoSetCode.SHR)
    self._parameters.append(mode)

class OppoSetInputSourceCommand(OppoSetCommand):
  def __init__(self, source: SetInputSource):
    super().__init__(OppoSetCode.SIS)
    self._parameters.append(source)

class OppoSetScreensaverCommand(OppoSetCommand):
  def __init__(self, mode: SetScreenSaverMode):
    super().__init__(OppoSetCode.SSA)
    self._parameters.append(mode)

class OppoSetAppModeCommand(OppoSetCommand):
  def __init__(self, mode: SetAppMode):
    super().__init__(OppoSetCode.APP)
    self._parameters.append(mode)

class OppoSetSacdPriorityCommand(OppoSetCommand):
  def __init__(self, mode: SetSacdPriority):
    super().__init__(OppoSetCode.SSD)
    self._parameters.append(mode)

class OppoSetSacdOutputModeCommand(OppoSetCommand):
  def __init__(self, mode: SetSacdOutputMode):
    super().__init__(OppoSetCode.SDP)
    self._parameters.append(mode)

class OppoSetFwdModeCommand(OppoSetCommand):
  def __init__(self, mode: Optional[SetSpeedMode]):
    super().__init__(OppoSetCode.FWD)
    if mode is not None:
      self._parameters.append(mode)

class OppoSetRevModeCommand(OppoSetCommand):
  def __init__(self, mode: Optional[SetSpeedMode]):
    super().__init__(OppoSetCode.REV)
    if mode is not None:
      self._parameters.append(mode)

class OppoSetTitleCommand(OppoSetCommand):
  def __init__(self, title_number: int):
    super().__init__(OppoSetCode.SRH)
    self._parameters.append("T"+str(title_number))  

class OppoSetChapterCommand(OppoSetCommand):
  def __init__(self, chapter_number: int):
    super().__init__(OppoSetCode.SRH)
    self._parameters.append("C"+str(chapter_number))  

class OppoSetChapterPositionCommand(OppoSetCommand):
  def __init__(self, position: timedelta):
    super().__init__(OppoSetCode.SRH)
    self._parameters.append("C " + strfdelta(position,"{H:02}:{M:02}:{S:02}"))

class OppoSetTitlePositionCommand(OppoSetCommand):
  def __init__(self, position: timedelta):
    super().__init__(OppoSetCode.SRH)
    self._parameters.append("T " + strfdelta(position,"{H:02}:{M:02}:{S:02}"))



import enum
from itertools import chain
from typing import Union

@enum.unique
class OppoRemoteCode(enum.Enum):
  """List of Oppo remote codes (i.e. mapped to the remote)"""
  POW = "POW"
  EJT = "EJT"
  PON = "PON"
  POF = "POF"
  DIM = "DIM"
  PUR = "PUR"
  VUP = "VUP"
  VDN = "VDN"
  MUT = "MUT"
  NU1 = "NU1"
  NU2 = "NU2"
  NU3 = "NU3"
  NU4 = "NU4"
  NU5 = "NU5"
  NU6 = "NU6"
  NU7 = "NU7"
  NU8 = "NU8"
  NU9 = "NU9"
  NU0 = "NU0"
  CLR = "CLR"
  GOT = "GOT"
  HOM = "HOM"
  PUP = "PUP"
  PDN = "PDN"
  OSD = "OSD"
  TTL = "TTL"
  MNU = "MNU"
  NUP = "NUP"
  NLT = "NLT"
  NRT = "NRT"
  NDN = "NDN"
  SEL = "SEL"
  SET = "SET"
  RET = "RET"
  RED = "RED"
  GRN = "GRN"
  BLU = "BLU"
  YLW = "YLW"
  STP = "STP"
  PLA = "PLA"
  PAU = "PAU"
  PRE = "PRE"
  REV = "REV"
  FWD = "FWD"
  NXT = "NXT"
  AUD = "AUD"
  SUB = "SUB"
  ANG = "ANG"
  ZOM = "ZOM"
  SAP = "SAP"
  ATB = "ATB"
  RPT = "RPT"
  PIP = "PIP"
  HDM = "HDM"
  SUH = "SUH"
  OPT = "OPT"
  M3D = "M3D"
  SEH = "SEH"
  HDR = "HDR"
  INH = "INH"
  RLH = "RLH"
  AVS = "AVS"
  GPA = "GPA"
  NOP = "NOP"
  SRC = "SRC"

@enum.unique
class OppoQueryCode(enum.Enum):
  """List of current state query codes"""
  QVM = "QVM"
  QPW = "QPW"
  QVR = "QVR"
  QVL = "QVL"
  QHD = "QHD"
  QPL = "QPL"
  QTK = "QTK"
  QCH = "QCH"
  QTE = "QTE"
  QTR = "QTR"
  QCE = "QCE"
  QCR = "QCR"
  QEL = "QEL"
  QRE = "QRE"
  QDT = "QDT"
  QAT = "QAT"
  QST = "QST"
  QSH = "QSH"
  QOP = "QOP"
  QRP = "QRP"
  QZM = "QZM"
  QHR = "QHR"
  Q3D = "Q3D"
  QHS = "QHS"
  QIS = "QIS"
  QCD = "QCD"
  QFT = "QFT"
  QFN = "QFN"
  QTN = "QTN"
  QTA = "QTA"
  QTP = "QTP"
  QDS = "QDS"
  QAR = "QAR"
  QDR = "QDR"  

@enum.unique
class OppoSetCode(enum.Enum):
  """Advanced set command code list"""
  SVM = "SVM"
  SHD = "SHD"
  SZM = "SZM"
  SVL = "SVL"
  SRP = "SRP"
  SRH = "SRH"
  DPL = "DPL"
  RST = "RST"
  SSH = "SSH"
  SOP = "SOP"
  STC = "STC"
  SHR = "SHR"
  SIS = "SIS"
  SSA = "SSA"
  APP = "APP"
  SSD = "SSD"
  SDP = "SDP"

@enum.unique
class OppoUpdateCode(enum.Enum):
  UPW = "UPW"
  UPL = "UPL"
  UVL = "UVL"
  UDT = "UDT"
  UAT = "UAT"
  UST = "UST"
  UIS = "UIS"
  U3D = "U3D"
  UAR = "UAR"
  UTC = "UTC"
  UVO = "UVO"

@enum.unique
class OppoResponseCode(enum.Enum):
  QC1 = "QC1"
  QC2 = "QC2"

OppoRemoteCodeType = Union[OppoRemoteCode, str]  
OppoQueryCodeType = Union[OppoRemoteCode, str]
OppoSetCodeType = Union[OppoSetCode, str]

OppoCode = enum.Enum('OppoCode', [(i.name, i.value) for i in chain(OppoRemoteCode, OppoQueryCode, OppoSetCode, OppoUpdateCode, OppoResponseCode)])
#OppoCode = Union[OppoRemoteCode, OppoQueryCode, OppoSetCode, OppoUpdateCode]
OppoCodeType = Union[OppoCode, str]



import logging
import asyncio
from typing import Optional, TYPE_CHECKING
from dataclasses import dataclass
from datetime import datetime, timedelta
from .const import *
from .command import *
from .response import *
from .codes import *
from .helpers import clamp

if TYPE_CHECKING:
    from .client import OppoClient

_LOGGER = logging.getLogger(__name__)

@dataclass(repr=True, eq=True)
class OppoPlaybackStatus:
  """Playback attributes for the Oppo device"""
  track: int = 0
  track_total: int = 0
  chapter: int = 0
  chapter_total: int = 0
  track_elapsed_time: timedelta = timedelta(seconds=0)
  track_remaining_time: timedelta = timedelta(seconds=0)
  track_duration: timedelta = timedelta(seconds=0)
  chapter_elapsed_time: timedelta = timedelta(seconds=0)
  chapter_remaining_time: timedelta = timedelta(seconds=0)
  chapter_duration: timedelta = timedelta(seconds=0)
  total_elapsed_time: timedelta = timedelta(seconds=0)
  total_remaining_time: timedelta = timedelta(seconds=0)
  total_duration: timedelta = timedelta(seconds=0)
  audio_type: str = ""
  subtitle_type: str = ""
  repeat_mode: RepeatMode = RepeatMode.OFF
  video_3d_status: Video3dStatus = Video3dStatus.UNKNOWN
  video_hdr_status: VideoHdrStatus = VideoHdrStatus.UNKNOWN
  aspect_ratio: str = ""
  media_file_format: str = ""
  media_file_name: str = ""
  track_name: str = ""
  track_album: str = ""
  track_performer: str = "" 
  rev_speed: SpeedMode = SpeedMode.NORMAL
  fwd_speed: SpeedMode = SpeedMode.NORMAL

class OppoDevice:
  """Represents a low-level Oppo device"""
  def __init__(self, client: 'OppoClient', mac_address: Optional[str] = None):
    self._client = client
    self._mac_address = mac_address
    self._client.add_event_handler(EVENT_MESSAGE_RECEIVED, self._on_message_received)
    self._client.add_event_handler(EVENT_DISCONNECTED, self._on_client_disconnected)

    self.power_status = PowerStatus.DISCONNECTED
    self.playback_status = PlayStatus.OFF
    self.firmware_version = ""
    self._reset_attributes() 
    self._state_events_enabled = True
    self._update_lock = asyncio.Lock()

  @property
  def mac_address(self) -> str:
    """The oppo device's MAC address"""
    return self._mac_address.upper()

  @property
  def is_playing(self) -> bool:
    """Indicates whether the device is playing"""
    return self.playback_status not in [
      PlayStatus.OFF, 
      PlayStatus.HOME_MENU, 
      PlayStatus.MEDIA_CENTER, 
      PlayStatus.SCREEN_SAVER, 
      PlayStatus.SETUP
    ]  

  @property
  def cddb_id(self) -> str:
    """The disc id for the currently loaded disc"""
    return self.cddb_id_1 + self.cddb_id_2

  @property
  def is_updating(self) -> bool:
    """Indicates whether we are currently updating the state"""
    return not self._state_events_enabled

  async def async_request_update(self):
    """Request the device to send a full state update"""
    try:
      #async with self._update_lock:
        #disable individual state events since we're doing a bunch at once
      self._state_events_enabled = False
      await self._client.async_event(EVENT_DEVICE_STATE_UPDATING, self)
      await self._client.async_send_command(OppoSetVerboseModeCommand(SetVerboseMode.VERBOSE))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QVM))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QPW))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QVR))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QVL))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QHD))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QPL))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QDT))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QSH))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QOP))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QZM))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QHR))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QIS))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QAR))
      await self._client.async_send_command(OppoQueryCdCommand())

      #request media-related updates
      await self.async_request_media_update(False)
    finally:
      #re-enable state events and send the updated event
      self._state_events_enabled = True
      await self._client.async_event(EVENT_DEVICE_STATE_UPDATED, self)

  async def async_request_media_update(self, suspend_events: bool = True):
    try:
      if self.is_playing:
        if suspend_events:
          self._state_events_enabled = False
          await self._client.async_event(EVENT_DEVICE_STATE_UPDATING, self)

        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTK))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QCH))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTE))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTR))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QCE))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QCR))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QEL))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QRE))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QAT))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QST))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QRP))   
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.Q3D))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QHS))   
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QFT))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QFN))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTN))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTA))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTP))
        await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QDS))
        self._calculate_duration()
    finally:
      if suspend_events:
        self._state_events_enabled = True
        await self._client.async_event(EVENT_DEVICE_STATE_UPDATED, self)        

  async def async_request_position_update(self):
    """Requests a playback position update"""
    try:
      #async with self._update_lock:      
      self._state_events_enabled = False
      await self._client.async_event(EVENT_DEVICE_STATE_UPDATING, self)
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTK))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QCH))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTE))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QTR))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QCE))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QCR))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QEL))
      await self._client.async_send_command(OppoQueryCommand(OppoQueryCode.QRE))
      self._calculate_duration()
      
    finally:
      self._state_events_enabled = True
      await self._client.async_event(EVENT_DEVICE_STATE_UPDATED, self)

  async def async_send_command(self, code: OppoRemoteCodeType):
    """Sends a remote command to the device"""
    await self._client.async_send_command(OppoRemoteCommand(code))

  async def async_set_input_source(self, source: SetInputSource):
    """Sets the input source"""
    await self._client.async_send_command(OppoSetInputSourceCommand(source))

  async def async_set_volume(self, level: int):
    """Sets the volume, accepts a level between 0 and 100"""
    await self._client.async_send_command(OppoSetVolumeLevelCommand(clamp(level, 0, 100)))

  async def async_seek_position(self, position_type: SetSearchMode, position: timedelta):
    """Seeks to a given position within either the chapter or title."""
    if position_type == SetSearchMode.CHAPTER:
      await self._client.async_send_command(OppoSetChapterPositionCommand(position))
    else:
      await self._client.async_send_command(OppoSetTitlePositionCommand(position))  

  async def async_repeat_mode(self, mode: SetRepeatMode):
    """Sets the volume, accepts a level between 0 and 100"""
    await self._client.async_send_command(OppoSetRepeatModeCommand(mode))

  def _reset_attributes(self):
    """Initializes/resets device attributes"""
    self.is_muted = False
    self.tray_status = TrayStatus.CLOSE
    self.volume = 0
    self.hdmi_mode = HdmiMode.UNKNOWN
    self.subtitle_shift = 0
    self.osd_position = 0
    self.input_source = InputSource.UNKNOWN
    self.zoom_mode = ZoomMode.UNKNOWN
    self.hdr_setting = HdrSetting.UNKNOWN
    self.disc_type = DiscType.UNKNOWN
    self.playback_status = PlayStatus.OFF
    self.cddb_id_1 = ""
    self.cddb_id_2 = ""
    self.last_update_at = datetime.utcnow()

    self.playback_attributes = OppoPlaybackStatus()

  def _calculate_duration(self):
    """Calculates the track/chapter/total duration (these are not provided by the device directly)"""
    pa = self.playback_attributes
    pa.track_duration = pa.track_elapsed_time + pa.track_remaining_time
    pa.chapter_duration = pa.chapter_elapsed_time + pa.chapter_remaining_time
    pa.total_duration = pa.total_elapsed_time + pa.total_remaining_time

  async def _on_client_disconnected(self):
    """Handles the client disconnected event"""
    self.power_status = PowerStatus.DISCONNECTED
    self._reset_attributes()

  async def _on_message_received(self, response: OppoResponse):
    """Handles message received events, updating state as needed"""
    if response.result == ResultCode.ERROR:
      _LOGGER.info(f"Invalid response {response}, ignoring")
      return

    if self._state_events_enabled:
      await self._client.async_event(EVENT_DEVICE_STATE_UPDATING, self)

    try:
      #handle various special messages first (affecting play status/power)
      if isinstance(response, OppoPowerResponse):
        await self._handle_power_response(response.status)
      elif isinstance(response, OppoUpdatePowerStatusResponse):
        await self._handle_power_response(response.status)
      elif isinstance(response, OppoPlayResponse):
        await self._handle_play_response(response.status)
      elif isinstance(response, OppoUpdatePlayStatusResponse):
        if response.play_status:
          await self._handle_play_response(response.play_status)
        await response.mutate_state(self)
      else:
        #otherwise, modify the state based on the response
        await response.mutate_state(self)

      #indicate when we last updated
      self.last_update_at = datetime.utcnow()

      if self._state_events_enabled:
        await self._client.async_event(EVENT_DEVICE_STATE_UPDATED, self)
    except:
      _LOGGER.warning("Error updating state.", exc_info=True)

  async def _handle_power_response(self, new_status: PowerStatus):
    """Handles power change events"""
    if self.power_status != new_status:
      self.power_status = new_status
      if self.power_status == PowerStatus.ON:
        #make sure that verbose mode is enabled
        await self._client.async_send_command(OppoSetVerboseModeCommand(SetVerboseMode.VERBOSE))
        #request an update of the state since it was OFF/DISCONNECTED
        await self.async_request_update()

  async def _handle_play_response(self, new_status: PlayStatus):
    """Handles play status change events"""
    if self.playback_status != new_status:
      self.playback_status = new_status
      if self.is_playing and self._state_events_enabled:
        await self.async_request_update()



"""Oppo UDP SDK"""

__version__ = "0.1.10"

from .codes import OppoRemoteCode, OppoRemoteCodeType
from .const import *
from .exceptions import *
from .command import *
from .response.enums import *
from .client import OppoClient
from .device import OppoDevice, OppoPlaybackStatus



import asyncio
from typing import AsyncIterator

class OppoStreamIterator:
  def __init__(self, reader: asyncio.StreamReader):
    self._reader = reader
  
  def __aiter__(self):
    return self
  
  async def __anext__(self):
    val = await self._reader.readuntil(b'\r')
    if val == b'':
      raise StopAsyncIteration
    return val    

async def CancellableAsyncIterator(async_iterator: AsyncIterator, cancellation_event: asyncio.Event) -> AsyncIterator:
  cancellation_task = asyncio.create_task(cancellation_event.wait())
  result_iter = async_iterator.__aiter__()
  while not cancellation_event.is_set():
    done, pending = await asyncio.wait(
      [cancellation_task, result_iter.__anext__()],
      return_when=asyncio.FIRST_COMPLETED
    )
    for done_task in done:
      if done_task == cancellation_task:
        for pending_task in pending:
          await pending_task
        break
      else:
        yield done_task.result()



from typing import List
from ..codes import OppoQueryCode, OppoQueryCodeType, OppoCode, OppoCodeType
from .command import OppoCommand

class OppoQueryCommand(OppoCommand):
  def __init__(self, code: OppoQueryCodeType, parameters: List[str] = [], response_codes: List[str] = []):      
      super().__init__(code, parameters, response_codes)

  def _translate(self, code: OppoQueryCodeType):
      if isinstance(code, str):
        return OppoCode(OppoQueryCode(code).value)
      return OppoCode(code.value)

class OppoQueryCdCommand(OppoQueryCommand):
  def __init__(self):
      super().__init__(OppoQueryCode.QCD, [], ["QC2"] )

class OppoQueryDirectoryCommand(OppoQueryCommand):
  def __init__(self, item: int):
      super().__init__(OppoQueryCode.QDR, [item])
      


from .enums import *
from .command import OppoCommand

from .query_command import OppoQueryCommand
from .query_command import OppoQueryCdCommand
from .query_command import OppoQueryDirectoryCommand

from .remote_command import OppoRemoteCommand

from .set_command import OppoSetCommand
from .set_command import OppoSetVerboseModeCommand
from .set_command import OppoSetHdmiModeCommand
from .set_command import OppoSetZoomModeCommand
from .set_command import OppoSetVolumeLevelCommand
from .set_command import OppoSetRepeatModeCommand
from .set_command import OppoSetSubtitleShiftCommand
from .set_command import OppoSetOsdPositionCommand
from .set_command import OppoSetTimeCodeModeCommand
from .set_command import OppoSetHdrModeCommand
from .set_command import OppoSetInputSourceCommand
from .set_command import OppoSetScreensaverCommand
from .set_command import OppoSetAppModeCommand
from .set_command import OppoSetSacdPriorityCommand
from .set_command import OppoSetSacdOutputModeCommand
from .set_command import OppoSetFwdModeCommand
from .set_command import OppoSetRevModeCommand
from .set_command import OppoSetTitleCommand
from .set_command import OppoSetChapterCommand
from .set_command import OppoSetChapterPositionCommand
from .set_command import OppoSetTitlePositionCommand




from typing import Any, Tuple, Type
from ..const import *
from .response import *
from .mutator import *

class ResponseMapping(NamedTuple):
  """Represents a combination of response and mutator which is used to handle a response"""
  response_type: Type
  mutator: OppoStateMutator

def get_response(message: bytes) -> OppoResponse:
  """Gets the response for a given message."""
  parsed = _parse_message(message)

  #if we don't have a code, and the parameters are on/off, assume power code
  if parsed.code == "" and len(parsed.parameters) > 0 and parsed.parameters[0] in ["ON","OFF"]:
    parsed = OppoParsedResponse("QPW", parsed.result, parsed.parameters)

  try:
    tcode = _translate_code(parsed.code)
    if tcode in _MAPPING:
      mapping =_MAPPING[tcode]
      return mapping.response_type(parsed, mapping.mutator, message)
  except:
    pass

  return OppoResponse(parsed, OppoNopMutator(), message) 

def _parse_message(message: bytes) -> OppoParsedResponse:
  """Parses a message received to identify the code, result, and parameters """
  decoded = message.decode()

  try:
    segments = decoded[:-1].split(" ")
    
    p_code = segments[0][1:]

    #must be wrong verbose mode, don't have the actual code
    if p_code in ["OK","ER"]:
      p_result = p_code
      p_code = ""
      p_parameters = segments[1:]
    #update codes don't have a result...
    elif p_code.startswith('U'):
      p_result = "OK"
      p_parameters = segments[1:]
    else:
      p_result = segments[1]
      p_parameters = segments[2:]
  except:
    p_code = ""
    p_result = "ER"
    p_parameters = [] 

  return OppoParsedResponse(
    code = p_code,
    result = p_result,
    parameters = p_parameters
  )

def _translate_code(code: OppoCodeType):
  try:
    if isinstance(code, str):
      return OppoCode(code)    
  except:
    _LOGGER.info(f'Unexpected code received: {code}')
    pass
  return code  

_MAPPING = {
  OppoCode.QPW: ResponseMapping(OppoPowerResponse, OppoSimpleDeviceMutator('status',ATTR_DEVICE_POWER_STATUS)),
  OppoCode.PON: ResponseMapping(OppoPowerResponse, OppoSimpleDeviceMutator('status',ATTR_DEVICE_POWER_STATUS)),
  OppoCode.POF: ResponseMapping(OppoPowerResponse, OppoSimpleDeviceMutator('status',ATTR_DEVICE_POWER_STATUS)),
  OppoCode.POW: ResponseMapping(OppoPowerResponse, OppoSimpleDeviceMutator('status',ATTR_DEVICE_POWER_STATUS)),
  OppoCode.UPW: ResponseMapping(OppoUpdatePowerStatusResponse, OppoNopMutator()),

  OppoCode.QVM: ResponseMapping(OppoIntResponse, OppoNopMutator()),
  OppoCode.SVM: ResponseMapping(OppoIntResponse, OppoNopMutator()),
  OppoCode.QVR: ResponseMapping(OppoStringResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_FIRMWARE_VERSION)),
  OppoCode.QVL: ResponseMapping(OppoVolumeLevelResponse, OppoVolumeLevelMutator()),  
  OppoCode.SVL: ResponseMapping(OppoVolumeLevelResponse, OppoVolumeLevelMutator()),  
  #TODO: there's an undocumented UVL that needs to be mapped
  OppoCode.MUT: ResponseMapping(OppoVolumeLevelResponse, OppoVolumeLevelMutator()),    
  OppoCode.VUP: ResponseMapping(OppoVolumeLevelResponse, OppoVolumeLevelMutator()),    
  OppoCode.VDN: ResponseMapping(OppoVolumeLevelResponse, OppoVolumeLevelMutator()),    
  OppoCode.QHD: ResponseMapping(OppoHdmiModeResponse, OppoSimpleDeviceMutator('mode', ATTR_DEVICE_HDMI_MODE)),
  OppoCode.SHD: ResponseMapping(OppoHdmiModeResponse, OppoSimpleDeviceMutator('mode', ATTR_DEVICE_HDMI_MODE)),
  OppoCode.QPL: ResponseMapping(OppoPlayResponse, OppoSimpleDeviceMutator('status',ATTR_DEVICE_PLAYBACK_STATUS)),
  OppoCode.UPL: ResponseMapping(OppoUpdatePlayStatusResponse, OppoUpdatePlayStatusMutator()),
  OppoCode.QTK: ResponseMapping(OppoCurrentTotalResponse, OppoTrackTotalMutator()),
  OppoCode.QCH: ResponseMapping(OppoCurrentTotalResponse, OppoChapterTotalMutator()),
  OppoCode.QDT: ResponseMapping(OppoDiscTypeResponse, OppoSimpleDeviceMutator('disc_type', ATTR_DEVICE_DISC_TYPE)),
  OppoCode.UDT: ResponseMapping(OppoUpdateDiscTypeResponse, OppoSimpleDeviceMutator('disc_type', ATTR_DEVICE_DISC_TYPE)),
  OppoCode.QSH: ResponseMapping(OppoIntResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_SUBTITLE_SHIFT)),
  OppoCode.SSH: ResponseMapping(OppoIntResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_SUBTITLE_SHIFT)),
  OppoCode.QOP: ResponseMapping(OppoIntResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_OSD_POSITION)),
  OppoCode.SOP: ResponseMapping(OppoIntResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_OSD_POSITION)),
  OppoCode.QZM: ResponseMapping(OppoRepeatModeResponse, OppoSimpleDeviceMutator('mode', ATTR_DEVICE_ZOOM_MODE)),
  OppoCode.SZM: ResponseMapping(OppoRepeatModeResponse, OppoSimpleDeviceMutator('mode', ATTR_DEVICE_ZOOM_MODE)),
  OppoCode.QHR: ResponseMapping(OppoHdrSettingResponse, OppoSimpleDeviceMutator('setting', ATTR_DEVICE_HDR_SETTING)),
  OppoCode.SHR: ResponseMapping(OppoHdrSettingResponse, OppoSimpleDeviceMutator('setting', ATTR_DEVICE_HDR_SETTING)),
  OppoCode.QIS: ResponseMapping(OppoInputSourceResponse, OppoSimpleDeviceMutator('source', ATTR_DEVICE_INPUT_SOURCE)),
  OppoCode.SIS: ResponseMapping(OppoInputSourceResponse, OppoSimpleDeviceMutator('source', ATTR_DEVICE_INPUT_SOURCE)),
  OppoCode.UIS: ResponseMapping(OppoInputSourceResponse, OppoSimpleDeviceMutator('source', ATTR_DEVICE_INPUT_SOURCE)),
  OppoCode.QC1: ResponseMapping(OppoStringResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_CDDB_ID_1)),
  OppoCode.QC2: ResponseMapping(OppoStringResponse, OppoSimpleDeviceMutator('value', ATTR_DEVICE_CDDB_ID_2)),
 
  OppoCode.QTE: ResponseMapping(OppoTimeResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TRACK_ELAPSED_TIME)),
  OppoCode.QTR: ResponseMapping(OppoTimeResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TRACK_REMAINING_TIME)),
  OppoCode.QCE: ResponseMapping(OppoTimeResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_CHAPTER_ELAPSED_TIME)),
  OppoCode.QCR: ResponseMapping(OppoTimeResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_CHAPTER_REMAINING_TIME)),
  OppoCode.QEL: ResponseMapping(OppoTimeResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TOTAL_ELAPSED_TIME)),
  OppoCode.QRE: ResponseMapping(OppoTimeResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TOTAL_REMAINING_TIME)),
  OppoCode.QAT: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_AUDIO_TYPE)),
  OppoCode.UAT: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_AUDIO_TYPE)),
  OppoCode.QST: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_SUBTITLE_TYPE)),
  OppoCode.UST: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_SUBTITLE_TYPE)),
  OppoCode.QRP: ResponseMapping(OppoRepeatModeResponse, OppoSimplePlaybackMutator('mode', ATTR_PLAYBACK_REPEAT_MODE)),
  OppoCode.SRP: ResponseMapping(OppoRepeatModeResponse, OppoSimplePlaybackMutator('mode', ATTR_PLAYBACK_REPEAT_MODE)),
  OppoCode.Q3D: ResponseMapping(OppoVideo3dStatusResponse, OppoSimplePlaybackMutator('status', ATTR_PLAYBACK_VIDEO_3D_STATUS)),
  OppoCode.U3D: ResponseMapping(OppoVideo3dStatusResponse, OppoSimplePlaybackMutator('status', ATTR_PLAYBACK_VIDEO_3D_STATUS)),
  OppoCode.QHS: ResponseMapping(OppoVideoHdrStatusResponse, OppoSimplePlaybackMutator('status', ATTR_PLAYBACK_VIDEO_HDR_STATUS)),
  OppoCode.QFT: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_MEDIA_FILE_FORMAT)),
  OppoCode.QFN: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_MEDIA_FILE_NAME)),
  OppoCode.QTN: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TRACK_NAME)),
  OppoCode.QTA: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TRACK_ALBUM)),
  OppoCode.QTP: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_TRACK_PERFORMER)),
  OppoCode.QAR: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_ASPECT_RATIO)),
  OppoCode.UAR: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_ASPECT_RATIO)),

  OppoCode.EJT: ResponseMapping(OppoTrayStatusResponse, OppoTrayStatusMutator()),
  OppoCode.REV: ResponseMapping(OppoSpeedModeResponse, OppoSimplePlaybackMutator('mode', ATTR_PLAYBACK_REV_SPEED)),
  OppoCode.FWD: ResponseMapping(OppoSpeedModeResponse, OppoSimplePlaybackMutator('mode', ATTR_PLAYBACK_FWD_SPEED)),
  OppoCode.FWD: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_CAMERA_ANGLE)),
  OppoCode.ATB: ResponseMapping(OppoStringResponse, OppoSimplePlaybackMutator('value', ATTR_PLAYBACK_ATB_REPLAY)),

  OppoCode.UTC: ResponseMapping(OppoUpdateTimeResponse, OppoUpdateTimeMutator()),
}



import asyncio
from asyncio.exceptions import InvalidStateError
import logging
from typing import Callable, DefaultDict, Dict, List, Optional, Tuple

from .codes import *
from .command import *
from .const import *
from .device import OppoDevice
from .exceptions import OppoCommandError, OppoInvalidStateError
from .response import *
from .states import OppoClientState
from .async_helpers import OppoStreamIterator, CancellableAsyncIterator

_LOGGER = logging.getLogger(__name__)

class OppoClient:
  """
  Creates a client that can communicate with an Oppo UDP-20x device.
  The command reference can be found at: http://download.oppodigital.com/UDP203/OPPO_UDP-20X_RS-232_and_IP_Control_Protocol.pdf
  
  Must supply the host/IP address and port (default 23).
  """
  def __init__(self, host_name: str, port_number: int = 23, mac_address: str = None, event_loop: Optional[asyncio.AbstractEventLoop] = None):
    self._host_name = host_name
    self._port_number = port_number
    self._loop = event_loop
    self._disconnect_requested = asyncio.Event()
    self._command_response_received = asyncio.Event()
    self._command_lock = asyncio.Lock()
    self._retries_since_last_connect = -1
    self._has_successful_connect = False   
    self._state = OppoClientState.INITIALIZING
    self._reader = None
    self._writer = None
    self._initialize_event_handlers()
    self._device = OppoDevice(self, mac_address)

  @property
  def state(self) -> OppoClientState:
    """Gets the state of the client"""
    return self._state

  @property
  def loop(self) -> asyncio.AbstractEventLoop:
    """Gets the asyncio event loop"""
    if self._loop is None:
      self._loop = asyncio.get_event_loop()
    return self._loop

  @property
  def device(self) -> OppoDevice:
    """Gets the device associated with this client"""
    return self._device

  @property
  def connected(self) -> bool:
    """Indicates whether the client is in a connected state"""
    return self._state not in [OppoClientState.DISCONNECTING, OppoClientState.DISCONNECTED]

  @property
  def available(self) -> bool:
    """Indicates whether the client is available for sending/receiving commands"""
    return self._state == OppoClientState.CONNECTED

  @property
  def event_handlers(self) -> Dict[str, List[Callable]]:
    return self._event_handlers

  async def async_event(self, event: str, *args, **kwargs):
    """Trigger event callbacks sequentially"""
    for cb in self.event_handlers[event]:
      asyncio.ensure_future(cb(*args, **kwargs), loop=self.loop)

  def add_event_handler(self, event: str, callback: Callable, disposable: bool = False):
    """Adds an event handler to an event"""
    if disposable:
      raise NotImplementedError('Support for disposable callbacks not yet implemented')
    self.event_handlers[event].append(callback)

  def remove_event_handler(self, event: str, callback: Callable):
    """Removes an event handler for an event"""
    try:
      self.event_handlers[event].remove(callable)
    except:
      _LOGGER.warn(f"could not remove event handler {event}-{callable}")

  def clear_event_handlers(self):
    """Clears all non-internal event handlers"""
    self._initialize_event_handlers()

  async def async_run_client(self):
    """Runs the client in an event loop including auto-reconnects if the client drops."""
    #reset the disconnect event
    self._disconnect_requested.clear()

    _LOGGER.info('Starting Oppo client')
    while not self._disconnect_requested.is_set():
      if self._retries_since_last_connect > MAX_RETRIES:
        _LOGGER.debug(f'Tried auto-reconnecting {MAX_RETRIES} times, giving up.')
        break
      try:
        await self._async_run_client()
      except OppoCommandError as err:
        _LOGGER.info(f'Error executing command {err}')
      except Exception as err:
        if not self._has_successful_connect:
          _LOGGER.warn(f'Unhandled exception on first connect attempt: {err}, disconnecting')
          break
        _LOGGER.info(f'Unhandled exception while running client: {err}, ignoring and restarting', exc_info=True)  
      finally:
        if not self._disconnect_requested.is_set():
          await self._set_state(OppoClientState.DROPPED)
          await self._set_state(OppoClientState.WAITING)
          _LOGGER.debug('Waiting before reconnecting')
          await asyncio.sleep(RETRY_INTERVAL)
        self._retries_since_last_connect += 1

    #initiate the disconnection            
    await self.disconnect()

  async def disconnect(self):
    """Disconnect and cleanup."""
    if not self._disconnect_requested.is_set():
      _LOGGER.info("Disconnecting")
      await self._set_state(OppoClientState.DISCONNECTING)         
      self._disconnect_requested.set()
      await self._disconnect()
      await self._set_state(OppoClientState.DISCONNECTED)

  async def test_connection(self) -> bool:
    if self._state != OppoClientState.INITIALIZING:
      raise OppoInvalidStateError
    
    reader = None
    writer = None

    try:
      await self._set_state(OppoClientState.CONNECTING)
      reader, writer = await asyncio.open_connection(self._host_name, self._port_number)
    except:
      return False
    finally:
      if writer:
        writer.close()
        await writer.wait_closed()
    
    return True

  async def async_send_command(self, command: OppoCommand):
    """Sends a command to the client"""
    try:
      await self._send_command(command)
    except asyncio.exceptions.TimeoutError:
      _LOGGER.warning("Timeout waiting for command response.")
      pass

  async def _set_state(self, new_state: OppoClientState) -> bool:
    """Indicate that the state changed and raise an event"""
    if self._state != new_state:
      old_state = self._state
      self._state = new_state
      await self.async_event(EVENT_STATE_CHANGED, old_state, new_state)
      return True
    return False

  def _initialize_event_handlers(self):
    """Initializes the internal event handlers"""
    self._event_handlers = DefaultDict(list)  # type: Dict[str, List[Callable]]
    self.add_event_handler(EVENT_STATE_CHANGED, self._on_state_change)
    self.add_event_handler(EVENT_COMMAND_RESPONSE, self._on_command_response)
    pass

  async def _on_state_change(self, old_state: OppoClientState, new_state: OppoClientState):
    """Handles the on state change event"""
    _LOGGER.debug(f'Client changed state: {old_state} to {new_state}')

    if new_state == OppoClientState.CONNECTED:
      await self.async_event(EVENT_CONNECTED, None)
    if new_state == OppoClientState.DISCONNECTED:
      await self.async_event(EVENT_DISCONNECTED, None)

  async def _on_command_response(self, response: OppoResponse):
    """Handles a command response received event"""
    _LOGGER.debug(f'Received command response: {response}')
#    async with self._command_lock:
    self._current_command = None
    self._command_response_received.set()

  async def _set_connected(self):
    """Marks the client as connected"""
    self._retries_since_last_connect = -1
    self._has_successful_connect = True
    await self._set_state(OppoClientState.CONNECTED)

  async def _async_run_client(self):
    """Run the client (internal)."""
    try:
      await self._set_state(OppoClientState.CONNECTING)
      reader, writer = await asyncio.open_connection(self._host_name, self._port_number)
      self._reader = reader
      self._writer = writer

      await self._set_connected()
      self._initialize_device()
      try:
        async for message in CancellableAsyncIterator(OppoStreamIterator(reader), self._disconnect_requested):
          await self._process_message(message)
      except RuntimeError as err:
        #do nothing if it's a StopAsyncIteration, we just stopped the iteration
        #as part of the disconnect
        if not isinstance(err.__cause__, StopAsyncIteration):
          raise
    finally:
      await self._disconnect()    

  async def _disconnect(self) -> None:
    """Disconnects the client (internal)"""
    if self._writer:
      self._writer.close()
      await self._writer.wait_closed()

  def _initialize_device(self):
    """Initializes the device (gets power status, and triggers a future initialization)"""
    async def _async_initialize_device():   
      await self.async_send_command(OppoQueryCommand(OppoQueryCode.QPW))

      #force an update if we are on...
      if self.device.power_status == PowerStatus.ON:
        await self.device.async_request_update()

      #indicate we're ready to go
      await self.async_event(EVENT_READY, self)
    
    asyncio.ensure_future(_async_initialize_device())

  async def _send_command(self, command: OppoCommand):
    """Sends a command to the client (internal)"""
    _LOGGER.debug(f'Sending command: {command}')
    async with self._command_lock:
      await self.async_event(EVENT_COMMAND_SENDING, command)
      self._current_command = command.expected_response_codes
      self._command_response_received.clear()
      if self._writer:
        #write the command to the stream
        self._writer.write(command.encode())  
        await self._writer.drain()
        await self.async_event(EVENT_COMMAND_SENT, command)
      await asyncio.wait_for(self._command_response_received.wait(), COMMAND_TIMEOUT)

  async def _process_message(self, message: bytes) -> OppoResponse:
    """Processes a message received from the Oppo server"""
    _LOGGER.debug(f'Received message: {message}')
    response = get_response(message)
    _LOGGER.debug(f'Parsed message: {response}')
    await self.async_event(EVENT_MESSAGE_RECEIVED, response)

    #if we're not in the right mode, we just need to assume that everything is good
    if response.raw_value[:3] in [b"@OK",b"@ER"] and self._current_command is not None:
      await self.async_event(EVENT_COMMAND_RESPONSE, response)
    #if this event paired to the current command, we need to indicate we've received
    #the response so that we can initiate more commands
    elif (
      self._current_command is not None and 
      isinstance(response.code, OppoCode) and 
      response.code.value in self._current_command
    ):
      await self.async_event(EVENT_COMMAND_RESPONSE, response)

    return response



from typing import List
from ..codes import *
from dataclasses import dataclass

@dataclass(repr=True, eq=True)
class OppoCommand:
  """Represents a command to an OppoDevice"""
  code: OppoCodeType
  _parameters: List[str]
  _response_codes: List[str]

  def __init__(self, code: OppoCodeType, parameters: List[str] = None, response_codes: List[str] = None):
    if parameters is None:
      parameters = []
    if response_codes is None:
      response_codes = []

    self.code = self._translate(code)
    self._parameters = parameters
    self._response_codes = response_codes + [self.code.value]

  def encode(self):
    params = ""
    if len(self._parameters) > 0:
      params = " " + " ".join(list(map(str, self._parameters)))
    return f"#{self.code.value}{params}\r".encode()

  @property
  def expected_response_codes(self):
    return self._response_codes

  def _translate(self, code: OppoCodeType):
    if isinstance(code, str):
      return OppoCode(code)
    return code





import abc
from datetime import timedelta
from typing import TYPE_CHECKING
import magicattr

from .enums import VolumeLevel

if TYPE_CHECKING:
  from ..device import OppoDevice
  from .response import (
    OppoResponse, 
    OppoVolumeLevelResponse, 
    OppoUpdatePlayStatusResponse,
    OppoUpdateTimeResponse,
    OppoCurrentTotalResponse
  )

class OppoStateMutator(metaclass=abc.ABCMeta):
  """Represents a mutator that can act on an OppoDevice to change its state"""
  def __init__(self) -> None:
    pass

  @abc.abstractmethod
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoResponse') -> None:
    """ Mutates the state of the device based on the response """
    pass

class OppoNopMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoResponse') -> None:
    pass

class OppoSimpleDeviceMutator(OppoStateMutator):
  def __init__(self, response_attr: str, device_attr: str) -> None:
    self.response_attr = response_attr
    self.device_attr = device_attr

  async def mutate_state(self, device: 'OppoDevice', response: 'OppoResponse') -> None:
    magicattr.set(device, self.device_attr, magicattr.get(response, self.response_attr))

class OppoSimplePlaybackMutator(OppoStateMutator):
  def __init__(self, response_attr: str, playback_attr: str) -> None:
    self.response_attr = response_attr
    self.playback_attr = playback_attr

  async def mutate_state(self, device: 'OppoDevice', response: 'OppoResponse') -> None:
    magicattr.set(device, f'playback_attributes.{self.playback_attr}', magicattr.get(response, self.response_attr))

class OppoVolumeLevelMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoVolumeLevelResponse') -> None:
    if isinstance(response.level, int):
      device.volume = response.level
      if response.level > 0:
        device.is_muted = False
    else:
      device.is_muted = response.level == VolumeLevel.MUTE 

class OppoUpdatePlayStatusMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoUpdatePlayStatusResponse') -> None:
    from ..device import OppoPlaybackStatus
    if response.tray_status:
      if device.tray_status != response.tray_status:
        device.tray_status = response.tray_status
        device.playback_attributes = OppoPlaybackStatus()
    if response.fwd_speed_mode:
      device.playback_attributes.fwd_speed = response.fwd_speed_mode
    if response.rev_speed_mode:
      device.playback_attributes.fwd_speed = response.rev_speed_mode

class OppoUpdateTimeMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoUpdateTimeResponse') -> None:
    needs_media_refresh = True if (
      device.playback_attributes.chapter != response.chapter_number or 
      device.playback_attributes.track != response.title_number
    ) else False

    if needs_media_refresh:
      if not device.is_updating:
        await device.async_request_media_update()
    else: 
      device.playback_attributes.chapter = response.chapter_number
      device.playback_attributes.track = response.title_number
      
      #figure out the deltas
      diff = timedelta(0)
      if response.time_type == "E":
        diff = response.time_value - device.playback_attributes.total_elapsed_time
      elif response.time_type == "T":
        diff = response.time_value - device.playback_attributes.track_elapsed_time
      elif response.time_type == "C":
        diff = response.time_value - device.playback_attributes.chapter_elapsed_time
      elif response.time_type == "R":
        diff = response.time_value - device.playback_attributes.total_remaining_time
      elif response.time_type == "X":
        diff = response.time_value - device.playback_attributes.track_remaining_time
      elif response.time_type == "K":
        diff = response.time_value - device.playback_attributes.chapter_remaining_time
      
      #apply the deltas  
      device.playback_attributes.total_elapsed_time += diff
      device.playback_attributes.track_elapsed_time += diff
      device.playback_attributes.chapter_elapsed_time += diff
      device.playback_attributes.total_remaining_time -= diff
      device.playback_attributes.track_remaining_time -= diff
      device.playback_attributes.chapter_remaining_time -= diff

      #if our remaining times go negative, reset them and request an
      #update to hopefully fix them up...
      if (
        device.playback_attributes.total_remaining_time <= timedelta(0) or
        device.playback_attributes.track_remaining_time <= timedelta(0) or
        device.playback_attributes.chapter_remaining_time <= timedelta(0)
      ):
        device.playback_attributes.total_remaining_time = timedelta(0)
        device.playback_attributes.track_remaining_time = timedelta(0)
        device.playback_attributes.chapter_remaining_time = timedelta(0)
        await device.async_request_position_update()

class OppoChapterTotalMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoCurrentTotalResponse') -> None:
    device.playback_attributes.chapter = response.current
    device.playback_attributes.chapter_total = response.total

class OppoTrackTotalMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoCurrentTotalResponse') -> None:
    device.playback_attributes.track = response.current
    device.playback_attributes.track_total = response.total

class OppoTrayStatusMutator(OppoStateMutator):
  async def mutate_state(self, device: 'OppoDevice', response: 'OppoResponse') -> None:
    from ..device import OppoPlaybackStatus
    if device.tray_status != response.status:
      device.tray_status = response.status
      device.playback_attributes = OppoPlaybackStatus()



class OppoException(Exception):
    """ Base class for all other custom exceptions """
    pass

class OppoCommandError(OppoException):
    """ Exception raised when there is an error processing a command """
    def __init__(self, message: str, code, reason, *args: object) -> None:
        super().__init__(*args)
        self.code = code
        self.reason = reason
        self.message = message
    
    def __str__(self) -> str:
        return f"There was an error while processing a command: Code={self.code}, Reason={self.reason}, Message={self.message}"

class OppoInvalidStateError(OppoException):
    """ Exception raised when an invalid operation is attempted """
    def __init__(self, message: str, *args: object) -> None:
        super().__init__(*args)
        self.message = message
    
    def __str__(self) -> str:
        return f"Invalid operation: {self.message}"




import enum
from typing import Union

@enum.unique
class ResultCode(enum.Enum):
  """Indicates whether the command was successful"""
  OK = "OK"
  ERROR = "ER"

@enum.unique
class PowerStatus(enum.Enum):
  """Indicate the current power status"""
  DISCONNECTED = "DISCONNECTED"
  ON = "ON"
  OFF = "OFF"

@enum.unique
class PlayStatus(enum.Enum):
  """Indicates the current play status"""
  OFF = "OFF"
  PLAY = "PLAY"
  PAUSE = "PAUSE"
  STOP = "STOP"
  STEP = "STEP"
  FAST_REVERSE = "FREV"
  FAST_FORWARD = "FFWD"
  SLOW_REVERSE = "SREV"
  SLOW_FORWARD = "SFWD"
  SETUP = "SETUP"
  HOME_MENU = "HOME_MENU"
  MEDIA_CENTER = "MEDIA CENTER"
  SCREEN_SAVER = "SCREEN SAVER"
  DISC_MENU = "DISC MENU"

@enum.unique
class VolumeLevel(enum.Enum):
  """Indicates the device volume level"""
  MUTE = "MUTE"
  UNMUTE = "UNMUTE"

  def __str__(self):
    return str(self.value)

VolumeLevelType = Union[VolumeLevel, int]  

@enum.unique
class TrayStatus(enum.Enum):
  """Indicates the tray status"""
  OPEN = "OPEN"
  CLOSE = "CLOSE"

@enum.unique
class HdmiMode(enum.Enum):
  """Indicates the HDMI output mode"""
  UNKNOWN = "UNKNOWN"
  AUTO = "AUTO"
  SRC = "Source Direct"
  UHD_AUTO = "UHD_AUTO"
  UHD24 = "UHD24"
  UHD50 = "UHD50"
  UHD60 = "UHD60"
  HD1080P_AUTO = "1080PAUTO"
  HD1080P24 = "1080P24"
  HD1080P50 = "1080P50"
  HD1080P60 = "1080P60"
  HD1080I50 = "1080I50"
  HD1080I60 = "1080I60"
  HD720P50 = "720P50"
  HD720P60 = "720P60"
  SD576P = "576P"
  SD576I = "576I"
  SD480P = "480P"
  SD480I = "480I"

@enum.unique
class ZoomMode(enum.Enum):
  """Indicate the zoom mode"""
  UNKNOWN = "UNKNOWN"
  OFF = "00"
  STRETCH = "01"
  FULL = "02"
  UNDERSCAN = "03"
  Z1D2X = "04"
  Z1D3X = "05"
  Z1D5X = "06"
  Z2X = "07"
  Z3X = "08"
  Z4X = "09"
  Z0D2X = "10"
  Z0D3X = "11"
  Z0D4X = "12"

@enum.unique
class HdrSetting(enum.Enum):
  """Indicate the HDR setting"""
  UNKNOWN = "UNKNOWN"
  AUTO = "Auto"
  ON = "On"
  OFF = "Off"
  STRIP_METADATA = "StripMetadata"

@enum.unique
class Video3dStatus(enum.Enum):
  """Indicate the 3D status for the playing video"""
  UNKNOWN = "UNKNOWN"
  V3D = "3D"
  V2D = "2D"

@enum.unique
class VideoHdrStatus(enum.Enum):
  """Indicate the HDR status for the playing video"""
  UNKNOWN = "UNKNOWN"
  HDR = "HDR"
  SDR = "SDR"
  DOV = "DOV"

@enum.unique
class InputSource(enum.Enum):
  """Indicates the input source"""
  UNKNOWN = "UNKNOWN"
  BLURAY = "0"
  HDMI_IN = "1"
  HDMI_ARC = "2"
  OPTICAL_IN = "3"
  COAX_IN = "4"
  USB_IN = "5"

@enum.unique
class DiscType(enum.Enum):
  """Indicates the type of disc"""
  UNKNOWN = "UNKNOW-DISC"
  BLURAY = "BD-MV"
  DVD_VIDEO = "DVD-VIDEO"
  DVD_AUDIO = "DVD-AUDIO"
  SACD = "SACD"
  CDDA = "CDDA"  
  UHD_BLURAY = "UHBD"
  NONE = "NO-DISC"
  DATA = "DATA"
  VCD2 = "VCD2"
  SVCD = "SVCD"

@enum.unique
class RepeatMode(enum.Enum):
  """Indicates the repeat mode"""
  OFF = "00"
  REPEAT_ONE = "01"
  REPEAT_CHAPTER = "02"
  REPEAT_ALL = "03"
  REPEAT_TITLE = "04"
  SHUFFLE = "05"
  RANDOM = "06"
  
@enum.unique
class SpeedMode(enum.Enum):
  """Indicates the speed mode for FWD/REV operations"""
  SLOW_1_32 = "1/32"
  SLOW_1_16 = "1/16"
  SLOW_1_8 = "1/8"
  SLOW_1_4 = "1/4"
  SLOW_1_2 = "1/2"
  NORMAL = "1"
  FAST_2 = "2"
  FAST_3 = "3"
  FAST_4 = "4"
  FAST_5 = "5"
  STEP = "STEP"

@enum.unique
class UpdatePlayStatus(enum.Enum):
  """Indicates the play status (from an update response)"""
  NO_DISC = "DISC"
  LOADING = "LOAD"
  TRAY_OPEN = "OPEN"
  TRAY_CLOSING = "CLOS"
  PLAY = "PLAY"
  PAUSED = "PAUS"
  STOPPED = "STOP"
  STP_FWD = "STPF"
  STP_REV = "STPR"
  FFWD_1 = "FFW1"
  FFWD_2 = "FFW2"
  FFWD_3 = "FFW3"
  FFWD_4 = "FFW4"
  FFWD_5 = "FFW5"
  FREV_1 = "FRV1"
  FREV_2 = "FRV2"
  FREV_3 = "FRV3"
  FREV_4 = "FRV4"
  FREV_5 = "FRV5"
  SFWD_1 = "SFW1"
  SFWD_2 = "SFW2"
  SFWD_3 = "SFW3"
  SFWD_4 = "SFW4"
  SFWD_5 = "SFW5"
  SREV_1 = "SRV1"
  SREV_2 = "SRV2"
  SREV_3 = "SRV3"
  SREV_4 = "SRV4"
  SREV_5 = "SRV5"
  HOME = "HOME"
  MEDIA_CENTER = "MCTR"
  SCREEN_SAVER = "SCVR"
  DISC_MENU = "MENU"

@enum.unique
class UpdateDiscType(enum.Enum):
  """Indicates the disc type (from an update response)"""
  UNKNOWN = "UNKW"
  BLURAY = "BDMV"
  DVD_VIDEO = "DVDV"
  DVD_AUDIO = "DVDA"
  SACD = "SACD"
  CDDA = "CDDA"  
  UHD_BLURAY = "UHBD"
  DATA = "DATA"
  VCD2 = "VCD2"
  SVCD = "SVCD"


from .enums import *

_UPDATE_PLAY_STATUS_TO_PLAY_STATUS = {
  UpdatePlayStatus.PLAY: PlayStatus.PLAY,
  UpdatePlayStatus.PAUSED: PlayStatus.PAUSE,
  UpdatePlayStatus.STOPPED: PlayStatus.STOP,
  UpdatePlayStatus.STP_FWD: PlayStatus.SLOW_FORWARD,
  UpdatePlayStatus.STP_REV: PlayStatus.SLOW_REVERSE,
  UpdatePlayStatus.FFWD_1: PlayStatus.FAST_FORWARD,
  UpdatePlayStatus.FFWD_2: PlayStatus.FAST_FORWARD,
  UpdatePlayStatus.FFWD_3: PlayStatus.FAST_FORWARD,
  UpdatePlayStatus.FFWD_4: PlayStatus.FAST_FORWARD,
  UpdatePlayStatus.FFWD_5: PlayStatus.FAST_FORWARD,
  UpdatePlayStatus.FREV_1: PlayStatus.FAST_REVERSE,
  UpdatePlayStatus.FREV_2: PlayStatus.FAST_REVERSE,
  UpdatePlayStatus.FREV_3: PlayStatus.FAST_REVERSE,
  UpdatePlayStatus.FREV_4: PlayStatus.FAST_REVERSE,
  UpdatePlayStatus.FREV_5: PlayStatus.FAST_REVERSE,
  UpdatePlayStatus.SFWD_1: PlayStatus.SLOW_FORWARD,
  UpdatePlayStatus.SFWD_2: PlayStatus.SLOW_FORWARD,
  UpdatePlayStatus.SFWD_3: PlayStatus.SLOW_FORWARD,
  UpdatePlayStatus.SFWD_4: PlayStatus.SLOW_FORWARD,
  UpdatePlayStatus.SFWD_5: PlayStatus.SLOW_FORWARD,
  UpdatePlayStatus.SREV_1: PlayStatus.SLOW_REVERSE,
  UpdatePlayStatus.SREV_2: PlayStatus.SLOW_REVERSE,
  UpdatePlayStatus.SREV_3: PlayStatus.SLOW_REVERSE,
  UpdatePlayStatus.SREV_4: PlayStatus.SLOW_REVERSE,
  UpdatePlayStatus.SREV_5: PlayStatus.SLOW_REVERSE,
  UpdatePlayStatus.HOME: PlayStatus.HOME_MENU,
  UpdatePlayStatus.MEDIA_CENTER: PlayStatus.MEDIA_CENTER,
  UpdatePlayStatus.SCREEN_SAVER: PlayStatus.SCREEN_SAVER,
  UpdatePlayStatus.DISC_MENU:  PlayStatus.DISC_MENU
}

_UPDATE_PLAY_STATUS_TO_TRAY_STATUS = {
  UpdatePlayStatus.NO_DISC: TrayStatus.CLOSE,
  UpdatePlayStatus.LOADING: TrayStatus.CLOSE,
  UpdatePlayStatus.TRAY_OPEN: TrayStatus.OPEN,
  UpdatePlayStatus.TRAY_CLOSING: TrayStatus.CLOSE
}

_UPDATE_PLAY_STATUS_TO_SPEED_MODE = {
  UpdatePlayStatus.STP_FWD: SpeedMode.STEP,
  UpdatePlayStatus.STP_REV: SpeedMode.STEP,
  UpdatePlayStatus.FFWD_1: SpeedMode.NORMAL,
  UpdatePlayStatus.FFWD_2: SpeedMode.FAST_2,
  UpdatePlayStatus.FFWD_3: SpeedMode.FAST_3,
  UpdatePlayStatus.FFWD_4: SpeedMode.FAST_4,
  UpdatePlayStatus.FFWD_5: SpeedMode.FAST_5,
  UpdatePlayStatus.FREV_1: SpeedMode.NORMAL,
  UpdatePlayStatus.FREV_2: SpeedMode.FAST_2,
  UpdatePlayStatus.FREV_3: SpeedMode.FAST_3,
  UpdatePlayStatus.FREV_4: SpeedMode.FAST_4,
  UpdatePlayStatus.FREV_5: SpeedMode.FAST_5,
  UpdatePlayStatus.SFWD_1: SpeedMode.SLOW_1_2,
  UpdatePlayStatus.SFWD_2: SpeedMode.SLOW_1_4,
  UpdatePlayStatus.SFWD_3: SpeedMode.SLOW_1_8,
  UpdatePlayStatus.SFWD_4: SpeedMode.SLOW_1_16,
  UpdatePlayStatus.SFWD_5: SpeedMode.SLOW_1_32,
  UpdatePlayStatus.SREV_1: SpeedMode.SLOW_1_2,
  UpdatePlayStatus.SREV_2: SpeedMode.SLOW_1_4,
  UpdatePlayStatus.SREV_3: SpeedMode.SLOW_1_8,
  UpdatePlayStatus.SREV_4: SpeedMode.SLOW_1_16,
  UpdatePlayStatus.SREV_5: SpeedMode.SLOW_1_32,
}

_UPDATE_DISC_TYPE_TO_DISC_TYPE = {
  UpdateDiscType.UNKNOWN: DiscType.UNKNOWN,
  UpdateDiscType.BLURAY: DiscType.BLURAY,
  UpdateDiscType.DVD_VIDEO: DiscType.DVD_VIDEO,
  UpdateDiscType.DVD_AUDIO: DiscType.DVD_AUDIO,
  UpdateDiscType.SACD: DiscType.SACD,
  UpdateDiscType.CDDA: DiscType.CDDA,
  UpdateDiscType.UHD_BLURAY: DiscType.UHD_BLURAY,
  UpdateDiscType.DATA: DiscType.DATA,
  UpdateDiscType.VCD2: DiscType.VCD2,
  UpdateDiscType.SVCD: DiscType.SVCD
}


from ..codes import OppoRemoteCode, OppoRemoteCodeType, OppoCode, OppoCodeType
from .command import OppoCommand

class OppoRemoteCommand(OppoCommand):
  def __init__(self, code: OppoRemoteCodeType):      
      super().__init__(code)

  def _translate(self, code: OppoRemoteCodeType):
      if isinstance(code, str):
        return OppoCode(OppoRemoteCode(code).value)
      return OppoCode(code.value)